@@ -36,18 +36,14 @@ const std::string Mata::Nfa::TYPE_NFA = "NFA";
 
 namespace {
     Simlib::Util::BinaryRelation compute_fw_direct_simulation(const Nfa& aut) {
-        Simlib::ExplicitLTS LTSforSimulation;
         Symbol maxSymbol = 0;
         const size_t state_num = aut.size();
+        Simlib::ExplicitLTS LTSforSimulation(state_num);
 
-        for (State stateFrom = 0; stateFrom < state_num; ++stateFrom) {
-            for (const Move &t : aut.get_moves_from(stateFrom)) {
-                for (State stateTo : t.targets) {
-                    LTSforSimulation.add_transition(stateFrom, t.symbol, stateTo);
-                }
-                if (t.symbol > maxSymbol) {
-                    maxSymbol = t.symbol;
-                }
+        for (const auto& transition : aut.delta) {
+            LTSforSimulation.add_transition(transition.src, transition.symb, transition.tgt);
+            if (transition.symb > maxSymbol) {
+                maxSymbol = transition.symb;
             }
         }
 
@@ -119,8 +115,8 @@ namespace {
                     }
 
                     // add the transition 'q_class_state-q_trans.symbol->representatives_class_states' at the end of transition list of transitions starting from q_class_state
-                    // as the q_trans.symbol should be largest symbol we saw (as we iterate trough getTransitionsFromState(q) which is ordered)
-                    result.delta[q_class_state].insert(Move(q_trans.symbol, representatives_class_states));
+                    // as the q_trans.symbol should be the largest symbol we saw (as we iterate trough getTransitionsFromState(q) which is ordered)
+                    result.delta.get_mutable_post(q_class_state).insert(Move(q_trans.symbol, representatives_class_states));
                 }
 
                 if (aut.final[q]) { // if q is final, then all states in its class are final => we make q_class_state final
@@ -228,7 +224,7 @@ namespace {
                     }
                 }
                 if (!new_state_trans_with_symbol.empty()) {
-                    trimmed_aut.delta[original_state_mapping.second].insert(new_state_trans_with_symbol);
+                    trimmed_aut.delta.get_mutable_post(original_state_mapping.second).insert(new_state_trans_with_symbol);
                 }
             }
         }
@@ -501,8 +497,8 @@ State Delta::find_max_state() {
 ///// Nfa structure related methods
 
 State Nfa::add_state() {
-    m_num_of_states = size() + 1;
-    return m_num_of_states - 1;
+    m_num_of_requested_states = size() + 1;
+    return m_num_of_requested_states - 1;
 }
 
 void Nfa::remove_epsilon(const Symbol epsilon)
@@ -532,22 +528,35 @@ StateSet Nfa::get_terminating_states() const
     return revert(*this).get_reachable_states();
 }
 
-void Nfa::trim()
+void Nfa::trim(StateToStateMap* state_map)
 {
-    *this = get_trimmed_automaton();
+    if (!state_map) {
+        StateToStateMap tmp_state_map{};
+        *this = get_trimmed_automaton(&tmp_state_map);
+    } else {
+        state_map->clear();
+        *this = get_trimmed_automaton(state_map);
+    }
 }
 
-Nfa Nfa::get_trimmed_automaton() {
+Nfa Nfa::get_trimmed_automaton(StateToStateMap* state_map) const {
     if (initial.empty() || final.empty()) { return Nfa{}; }
 
+    StateToStateMap tmp_state_map{};
+    if (!state_map) {
+        state_map = &tmp_state_map;
+    }
+    state_map->clear();
+
     const StateSet original_useful_states{ get_useful_states() };
-    StateToStateMap original_to_new_states_map{ original_useful_states.size() };
+    state_map->reserve(original_useful_states.size());
+
     size_t new_state_num{ 0 };
     for (const State original_state: original_useful_states) {
-        original_to_new_states_map[original_state] = new_state_num;
+        (*state_map)[original_state] = new_state_num;
         ++new_state_num;
     }
-    return create_trimmed_aut(*this, original_to_new_states_map);
+    return create_trimmed_aut(*this, *state_map);
 }
 
 StateSet Nfa::get_useful_states() const
@@ -740,9 +749,9 @@ Nfa Mata::Nfa::revert(const Nfa& aut) {
     Nfa result;
     result.clear();
 
-    result.add_state(aut.m_num_of_states);
+    const size_t num_of_states{ aut.size() };
+    result.add_state(num_of_states - 1);
 
-    const size_t num_of_states{aut.size() };
     for (State sourceState{ 0 }; sourceState < num_of_states; ++sourceState) {
         for (const Move &transition: aut.delta[sourceState]) {
             for (const State targetState: transition.targets) {
@@ -1054,7 +1063,7 @@ TransSequence Nfa::get_transitions_to(State state_to) const {
             const auto& symbol_states_to{ symbol_transitions.targets };
             const auto target_state{ symbol_states_to.find(state_to) };
             if (target_state != symbol_states_to.end()) {
-                transitions_to_state.push_back({ state_from, symbol_transitions.symbol, state_to });
+                transitions_to_state.emplace_back( state_from, symbol_transitions.symbol, state_to );
             }
         }
     }
@@ -1124,7 +1133,7 @@ Nfa Mata::Nfa::uni(const Nfa &lhs, const Nfa &rhs) {
                 transitionFromUnionState.insert(thisStateToUnionState[stateTo]);
             }
 
-            unionAutomaton.delta[unionState].insert(transitionFromUnionState);
+            unionAutomaton.delta.get_mutable_post(unionState).insert(transitionFromUnionState);
         }
     }
 
@@ -1154,27 +1163,44 @@ Simlib::Util::BinaryRelation Mata::Nfa::Algorithms::compute_relation(const Nfa&
     }
 }
 
-Nfa Mata::Nfa::reduce(const Nfa &aut, StateToStateMap *state_map, const StringMap& params) {
-    Nfa result{};
-
+Nfa Mata::Nfa::reduce(const Nfa &aut, bool trim_input, StateToStateMap *state_map, const StringMap& params) {
     if (!haskey(params, "algorithm")) {
         throw std::runtime_error(std::to_string(__func__) +
                                  " requires setting the \"algorithm\" key in the \"params\" argument; "
                                  "received: " + std::to_string(params));
     }
 
+    Nfa aut_to_reduce{ aut };
+    StateToStateMap trimmed_state_map{};
+    if (trim_input) {
+        aut_to_reduce.trim(&trimmed_state_map);
+    }
+
+    Nfa result;
+    std::unordered_map<State,State> reduced_state_map;
     const std::string& algorithm = params.at("algorithm");
     if ("simulation" == algorithm) {
-        if (state_map == nullptr) {
-            std::unordered_map<State,State> tmp_state_map;
-            return reduce_size_by_simulation(aut, tmp_state_map);
-        } else {
-            return reduce_size_by_simulation(aut, *state_map);
-        }
+        result = reduce_size_by_simulation(aut_to_reduce, reduced_state_map);
     } else {
         throw std::runtime_error(std::to_string(__func__) +
                                  " received an unknown value of the \"algorithm\" key: " + algorithm);
     }
+
+    if (state_map) {
+        state_map->clear();
+        if (trim_input) {
+            for (const auto& trimmed_mapping : trimmed_state_map) {
+                const auto reduced_mapping{ reduced_state_map.find(trimmed_mapping.second) };
+                if (reduced_mapping != reduced_state_map.end()) {
+                    (*state_map)[trimmed_mapping.first] = reduced_mapping->second;
+                }
+            }
+        } else { // Input has not been trimmed, the reduced state map is the actual input to result state map.
+            *state_map = reduced_state_map;
+        }
+    }
+
+    return result;
 }
 
 Nfa Mata::Nfa::determinize(
@@ -1248,7 +1274,7 @@ Nfa Mata::Nfa::determinize(
                 }
                 worklist.emplace_back(std::make_pair(Tid, T));
             }
-            result.delta[Sid].insert(Move(currentSymbol, Tid));
+            result.delta.get_mutable_post(Sid).insert(Move(currentSymbol, Tid));
         }
     }
 
@@ -1257,6 +1283,7 @@ Nfa Mata::Nfa::determinize(
     return result;
 }
 
+// TODO this function should the same thing as the one taking IntermediateAut or be deleted
 Nfa Mata::Nfa::construct(
         const Mata::Parser::ParsedSection&   parsec,
         Alphabet*                            alphabet,
@@ -1359,10 +1386,9 @@ Nfa Mata::Nfa::construct(
                                  Mata::Nfa::TYPE_NFA + "\"");
     }
 
-    bool remove_state_map = false;
+    StringToStateMap tmp_state_map;
     if (nullptr == state_map) {
-        state_map = new StringToStateMap();
-        remove_state_map = true;
+        state_map = &tmp_state_map;
     }
 
     // a lambda for translating state names to identifiers
@@ -1376,32 +1402,16 @@ Nfa Mata::Nfa::construct(
         }
     };
 
-    // a lambda for cleanup
-    auto clean_up = [&]() {
-        if (remove_state_map) { delete state_map; }
-    };
-
     for (const auto& str : inter_aut.initial_formula.collect_node_names())
     {
         State state = get_state_name(str);
         aut.initial.add(state);
     }
 
-    const auto final_states = inter_aut.are_final_states_conjunction_of_negation() ?
-            inter_aut.get_positive_finals() : inter_aut.final_formula.collect_node_names();
-    for (const auto& str : final_states)
-    {
-        State state = get_state_name(str);
-        aut.final.add(state);
-    }
-
     for (const auto& trans : inter_aut.transitions)
     {
         if (trans.second.children.size() != 2)
         {
-            // clean up
-            clean_up();
-
             if (trans.second.children.size() == 1)
             {
                 throw std::runtime_error("Epsilon transitions not supported");
@@ -1419,8 +1429,29 @@ Nfa Mata::Nfa::construct(
         aut.delta.add(src_state, symbol, tgt_state);
     }
 
-    // do the dishes and take out garbage
-    clean_up();
+    std::unordered_set<std::string> final_formula_nodes;
+    if (!(inter_aut.final_formula.node.is_constant())) {
+        // we do not want to parse true/false (constant) as a state so we do not collect it
+        final_formula_nodes = inter_aut.final_formula.collect_node_names();
+    }
+    // for constant true, we will pretend that final nodes are negated with empty final_formula_nodes
+    bool final_nodes_are_negated = (inter_aut.final_formula.node.is_true() || inter_aut.are_final_states_conjunction_of_negation());
+
+    if (final_nodes_are_negated) {
+        // we add all states NOT in final_formula_nodes to final states
+        for (const auto &state_name_and_id : *state_map) {
+            if (!final_formula_nodes.count(state_name_and_id.first)) {
+                aut.final.add(state_name_and_id.second);
+            }
+        }
+    } else {
+        // we add all states in final_formula_nodes to final states
+        for (const auto& str : final_formula_nodes)
+        {
+            State state = get_state_name(str);
+            aut.final.add(state);
+        }
+    }
 
     return aut;
 } // construct }}}
@@ -1568,7 +1599,7 @@ void Nfa::unify_final() {
     final.add(new_final_state);
 }
 
-void Nfa::fill_alphabet(const Mata::Nfa::Nfa& nfa, OnTheFlyAlphabet& alphabet) {
+void Mata::Nfa::fill_alphabet(const Mata::Nfa::Nfa& nfa, OnTheFlyAlphabet& alphabet) {
     size_t nfa_num_of_states{nfa.size() };
     for (Mata::Nfa::State state{ 0 }; state < nfa_num_of_states; ++state) {
         for (const auto state_transitions: nfa.delta) {
@@ -1577,6 +1608,7 @@ void Nfa::fill_alphabet(const Mata::Nfa::Nfa& nfa, OnTheFlyAlphabet& alphabet) {
         }
     }
 }
+
 void Nfa::add_symbols_to(OnTheFlyAlphabet& alphabet) {
     size_t aut_num_of_states{size() };
     for (Mata::Nfa::State state{ 0 }; state < aut_num_of_states; ++state) {
@@ -1586,3 +1618,47 @@ void Nfa::add_symbols_to(OnTheFlyAlphabet& alphabet) {
         }
     }
 }
+
+Mata::OnTheFlyAlphabet Mata::Nfa::create_alphabet(const ConstAutRefSequence& nfas) {
+    Mata::OnTheFlyAlphabet alphabet{};
+    for (const auto& nfa: nfas) {
+        fill_alphabet(nfa, alphabet);
+    }
+    return alphabet;
+}
+
+Mata::OnTheFlyAlphabet Mata::Nfa::create_alphabet(const AutRefSequence& nfas) {
+    Mata::OnTheFlyAlphabet alphabet{};
+    for (const auto& nfa: nfas) {
+        fill_alphabet(nfa, alphabet);
+    }
+    return alphabet;
+}
+
+Mata::OnTheFlyAlphabet Mata::Nfa::create_alphabet(const ConstAutPtrSequence& nfas) {
+    Mata::OnTheFlyAlphabet alphabet{};
+    for (const Mata::Nfa::Nfa* const nfa: nfas) {
+        fill_alphabet(*nfa, alphabet);
+    }
+    return alphabet;
+}
+
+Mata::OnTheFlyAlphabet Mata::Nfa::create_alphabet(const AutPtrSequence& nfas) {
+    Mata::OnTheFlyAlphabet alphabet{};
+    for (const Mata::Nfa::Nfa* const nfa: nfas) {
+        fill_alphabet(*nfa, alphabet);
+    }
+    return alphabet;
+}
+
+Nfa Mata::Nfa::create_empty_string_nfa() {
+    return Nfa{ 1, StateSet{ 0 }, StateSet{ 0 } };
+}
+
+Nfa Mata::Nfa::create_sigma_star_nfa(Mata::Alphabet* alphabet) {
+    Nfa nfa{ 1, StateSet{ 0 }, StateSet{ 0 }, alphabet };
+    for (const Mata::Symbol& symbol : alphabet->get_alphabet_symbols()) {
+        nfa.delta.add(0, symbol, 0);
+    }
+    return nfa;
+}
