@@ -127,7 +127,7 @@ TEST_CASE("Mata::Nfa::IntAlphabet") {
     CHECK(different_alphabet.is_equal(&different_alphabet));
 }
 
-TEST_CASE("Mata::Nfa::OnTheFlyAlphabet::from_nfas()") {
+TEST_CASE("Mata::Nfa::create_alphabet()") {
     Nfa a{1};
     a.delta.add(0, 'a', 0);
 
@@ -137,13 +137,13 @@ TEST_CASE("Mata::Nfa::OnTheFlyAlphabet::from_nfas()") {
     Nfa c{1};
     b.delta.add(0, 'c', 0);
 
-    auto alphabet{ Nfa::from_nfas(a, b, c) };
+    auto alphabet{Mata::Nfa::create_alphabet(a, b, c) };
 
     auto symbols{alphabet.get_alphabet_symbols() };
     CHECK(symbols == Mata::Util::OrdVector<Symbol>{ 'c', 'b', 'a' });
 
-    //OnTheFlyAlphabet::from_nfas(1, 3, 4); // Will not compile: '1', '3', '4' are not of the required type.
-    //OnTheFlyAlphabet::from_nfas(a, b, 4); // Will not compile: '4' is not of the required type.
+    //Mata::Nfa::create_alphabet(1, 3, 4); // Will not compile: '1', '3', '4' are not of the required type.
+    //Mata::Nfa::create_alphabet(a, b, 4); // Will not compile: '4' is not of the required type.
 }
 
 TEST_CASE("Mata::Nfa::OnTheFlyAlphabet::add_symbols_from()") {
@@ -364,7 +364,7 @@ TEST_CASE("Mata::Nfa::union_norename()")
 
 		OnTheFlyAlphabet alph{"a", "b"};
 		StringMap params;
-		params["algo"] = "antichains";
+		params["algorithm"] = "antichains";
 
 		REQUIRE(is_included(a, res, &alph, params));
 		REQUIRE(is_included(b, res, &alph, params));
@@ -380,7 +380,7 @@ TEST_CASE("Mata::Nfa::union_norename()")
 
 		OnTheFlyAlphabet alph{"a", "b"};
 		StringMap params;
-		params["algo"] = "antichains";
+		params["algorithm"] = "antichains";
 
 		REQUIRE(is_included(a, res, &alph, params));
 		REQUIRE(is_included(res, a, &alph, params));
@@ -627,6 +627,26 @@ TEST_CASE("Mata::Nfa::determinize()")
         REQUIRE(result.final[subset_map[{2}]]);
 		REQUIRE(result.delta.contains(subset_map[{1}], 'a', subset_map[{2}]));
 	}
+
+    SECTION("This broke Delta when delta[q] could cause re-allocation of post")
+    {
+        Nfa x{};
+        x.initial.add(0);
+        x.final.add(4);
+        x.delta.add(0, 1, 3);
+        x.delta.add(3, 1, 3);
+        x.delta.add(3, 2, 3);
+        x.delta.add(3, 0, 1);
+        x.delta.add(1, 1, 1);
+        x.delta.add(1, 2, 1);
+        x.delta.add(1, 0, 2);
+        x.delta.add(2, 0, 2);
+        x.delta.add(2, 1, 2);
+        x.delta.add(2, 2, 2);
+        x.delta.add(2, 0, 4);
+        OnTheFlyAlphabet alphabet{};
+        auto complement_result{determinize(x)};
+    }
 } // }}}
 
 TEST_CASE("Mata::Nfa::minimize() for profiling", "[.profiling],[minimize]") {
@@ -960,6 +980,61 @@ TEST_CASE("Mata::Nfa::construct() from IntermediateAut correct calls")
         REQUIRE(!is_in_lang(aut, encode_word(symbol_map, {"a1", "a2", "a3", "a4"})));
         REQUIRE(is_in_lang(aut, encode_word(symbol_map, {"a1", "a2", "a3", "a5", "a7"})));
     }
+
+    SECTION("construct - final states given as true")
+    {
+        std::string file =
+                "@NFA-bits\n"
+                "%Alphabet-auto\n"
+                "%Initial q0 q8\n"
+                "%Final true\n"
+                "q0 a1 q1\n"
+                "q1 a2 q2\n"
+                "q2 a3 q3\n"
+                "q2 a4 q4\n"
+                "q3 a5 q5\n"
+                "q3 a6 q6\n"
+                "q5 a7 q7\n";
+
+        const auto auts = Mata::IntermediateAut::parse_from_mf(parse_mf(file));
+        inter_aut = auts[0];
+
+		Mata::Nfa::StringToStateMap state_map;
+        construct(&aut, inter_aut, &symbol_map, &state_map);
+        CHECK(aut.final.size() == 9);
+		CHECK(aut.final[state_map.at("0")]);
+		CHECK(aut.final[state_map.at("1")]);
+		CHECK(aut.final[state_map.at("2")]);
+		CHECK(aut.final[state_map.at("3")]);
+		CHECK(aut.final[state_map.at("4")]);
+		CHECK(aut.final[state_map.at("5")]);
+		CHECK(aut.final[state_map.at("6")]);
+		CHECK(aut.final[state_map.at("7")]);
+		CHECK(aut.final[state_map.at("8")]);
+    }
+
+    SECTION("construct - final states given as false")
+    {
+        std::string file =
+                "@NFA-bits\n"
+                "%Alphabet-auto\n"
+                "%Initial q0 q8\n"
+                "%Final false\n"
+                "q0 a1 q1\n"
+                "q1 a2 q2\n"
+                "q2 a3 q3\n"
+                "q2 a4 q4\n"
+                "q3 a5 q5\n"
+                "q3 a6 q6\n"
+                "q5 a7 q7\n";
+
+        const auto auts = Mata::IntermediateAut::parse_from_mf(parse_mf(file));
+        inter_aut = auts[0];
+
+		Mata::Nfa::StringToStateMap state_map;
+        construct(&aut, inter_aut, &symbol_map, &state_map);
+        CHECK(aut.final.size() == 0);
+    }
 } // }}}
 
 /*
@@ -1286,7 +1361,7 @@ TEST_CASE("Mata::Nfa::is_universal()")
 		OnTheFlyAlphabet alph{};
 
 		for (const auto& algo : ALGORITHMS) {
-			params["algo"] = algo;
+			params["algorithm"] = algo;
 			bool is_univ = is_universal(aut, alph, params);
 
 			REQUIRE(!is_univ);
@@ -1300,7 +1375,7 @@ TEST_CASE("Mata::Nfa::is_universal()")
 		aut.final = {1};
 
 		for (const auto& algo : ALGORITHMS) {
-			params["algo"] = algo;
+			params["algorithm"] = algo;
 			bool is_univ = is_universal(aut, alph, &cex, params);
 
 			REQUIRE(is_univ);
@@ -1315,7 +1390,7 @@ TEST_CASE("Mata::Nfa::is_universal()")
 		aut.final = {1};
 
 		for (const auto& algo : ALGORITHMS) {
-			params["algo"] = algo;
+			params["algorithm"] = algo;
 			bool is_univ = is_universal(aut, alph, &cex, params);
 
 			REQUIRE(!is_univ);
@@ -1334,7 +1409,7 @@ TEST_CASE("Mata::Nfa::is_universal()")
 		aut.delta.add(2, alph["b"], 2);
 
 		for (const auto& algo : ALGORITHMS) {
-			params["algo"] = algo;
+			params["algorithm"] = algo;
 			bool is_univ = is_universal(aut, alph, params);
 
 			REQUIRE(!is_univ);
@@ -1351,7 +1426,7 @@ TEST_CASE("Mata::Nfa::is_universal()")
 		aut.delta.add(2, alph["b"], 2);
 
 		for (const auto& algo : ALGORITHMS) {
-			params["algo"] = algo;
+			params["algorithm"] = algo;
 			bool is_univ = is_universal(aut, alph, params);
 
 			REQUIRE(!is_univ);
@@ -1368,7 +1443,7 @@ TEST_CASE("Mata::Nfa::is_universal()")
 		aut.delta.add(1, alph["b"], 1);
 
 		for (const auto& algo : ALGORITHMS) {
-			params["algo"] = algo;
+			params["algorithm"] = algo;
 			bool is_univ = is_universal(aut, alph, params);
 
 			REQUIRE(is_univ);
@@ -1393,7 +1468,7 @@ TEST_CASE("Mata::Nfa::is_universal()")
 		aut.delta.add(5, alph["b"], 5);
 
 		for (const auto& algo : ALGORITHMS) {
-			params["algo"] = algo;
+			params["algorithm"] = algo;
 			bool is_univ = is_universal(aut, alph, &cex, params);
 
 			REQUIRE(!is_univ);
@@ -1421,7 +1496,7 @@ TEST_CASE("Mata::Nfa::is_universal()")
 		aut.delta.add(4, alph["b"], 4);
 
 		for (const auto& algo : ALGORITHMS) {
-			params["algo"] = algo;
+			params["algorithm"] = algo;
 			bool is_univ = is_universal(aut, alph, &cex, params);
 
 			REQUIRE(is_univ);
@@ -1445,7 +1520,7 @@ TEST_CASE("Mata::Nfa::is_universal()")
 		aut.delta.add(4, alph["b"], 3);
 
 		for (const auto& algo : ALGORITHMS) {
-			params["algo"] = algo;
+			params["algorithm"] = algo;
 			bool is_univ = is_universal(aut, alph, &cex, params);
 
 			REQUIRE(is_univ);
@@ -1461,7 +1536,7 @@ TEST_CASE("Mata::Nfa::is_universal()")
 		aut.delta.add(1, alph["a"], 1);
 
 		for (const auto& algo : ALGORITHMS) {
-			params["algo"] = algo;
+			params["algorithm"] = algo;
 			bool is_univ = is_universal(aut, alph, &cex, params);
 
 			REQUIRE(is_univ);
@@ -1479,7 +1554,7 @@ TEST_CASE("Mata::Nfa::is_universal()")
 	SECTION("wrong parameters 2")
 	{
 		OnTheFlyAlphabet alph{};
-		params["algo"] = "foo";
+		params["algorithm"] = "foo";
 
 		CHECK_THROWS_WITH(is_universal(aut, alph, params),
 			Catch::Contains("received an unknown value"));
@@ -1503,7 +1578,7 @@ TEST_CASE("Mata::Nfa::is_included()")
 		OnTheFlyAlphabet alph{};
 
 		for (const auto& algo : ALGORITHMS) {
-			params["algo"] = algo;
+			params["algorithm"] = algo;
 			bool is_incl = is_included(smaller, bigger, &alph, params);
 			CHECK(is_incl);
 
@@ -1519,7 +1594,7 @@ TEST_CASE("Mata::Nfa::is_included()")
 		bigger.final = {1};
 
 		for (const auto& algo : ALGORITHMS) {
-			params["algo"] = algo;
+			params["algorithm"] = algo;
 			bool is_incl = is_included(smaller, bigger, &cex, &alph, params);
             CHECK(is_incl);
 
@@ -1537,7 +1612,7 @@ TEST_CASE("Mata::Nfa::is_included()")
 		bigger.final = {11};
 
 		for (const auto& algo : ALGORITHMS) {
-			params["algo"] = algo;
+			params["algorithm"] = algo;
 			bool is_incl = is_included(smaller, bigger, &cex, &alph, params);
             CHECK(is_incl);
 
@@ -1553,7 +1628,7 @@ TEST_CASE("Mata::Nfa::is_included()")
 		smaller.final = {1};
 
 		for (const auto& algo : ALGORITHMS) {
-			params["algo"] = algo;
+			params["algorithm"] = algo;
 			bool is_incl = is_included(smaller, bigger, &cex, &alph, params);
 
 			REQUIRE(!is_incl);
@@ -1579,7 +1654,7 @@ TEST_CASE("Mata::Nfa::is_included()")
 		bigger.delta.add(11, alph["b"], 11);
 
 		for (const auto& algo : ALGORITHMS) {
-			params["algo"] = algo;
+			params["algorithm"] = algo;
 			bool is_incl = is_included(smaller, bigger, &alph, params);
 			REQUIRE(is_incl);
 
@@ -1602,7 +1677,7 @@ TEST_CASE("Mata::Nfa::is_included()")
 		bigger.delta.add(12, alph["b"], 12);
 
 		for (const auto& algo : ALGORITHMS) {
-			params["algo"] = algo;
+			params["algorithm"] = algo;
 
 			bool is_incl = is_included(smaller, bigger, &cex, &alph, params);
 
@@ -1642,7 +1717,7 @@ TEST_CASE("Mata::Nfa::is_included()")
 		bigger.delta.add(15, alph["b"], 15);
 
 		for (const auto& algo : ALGORITHMS) {
-			params["algo"] = algo;
+			params["algorithm"] = algo;
 			bool is_incl = is_included(smaller, bigger, &cex, &alph, params);
 			REQUIRE(!is_incl);
 
@@ -1677,7 +1752,7 @@ TEST_CASE("Mata::Nfa::is_included()")
 	SECTION("wrong parameters 2")
 	{
         OnTheFlyAlphabet alph{};
-		params["algo"] = "foo";
+		params["algorithm"] = "foo";
 
 		CHECK_THROWS_WITH(is_included(smaller, bigger, &alph, params),
 			Catch::Contains("received an unknown value"));
@@ -1702,7 +1777,7 @@ TEST_CASE("Mata::Nfa::are_equivalent")
         OnTheFlyAlphabet alph{};
 
         for (const auto& algo : ALGORITHMS) {
-            params["algo"] = algo;
+            params["algorithm"] = algo;
 
             CHECK(are_equivalent(smaller, bigger, &alph, params));
             CHECK(are_equivalent(smaller, bigger, params));
@@ -1721,7 +1796,7 @@ TEST_CASE("Mata::Nfa::are_equivalent")
         bigger.final = {1};
 
         for (const auto& algo : ALGORITHMS) {
-            params["algo"] = algo;
+            params["algorithm"] = algo;
 
             CHECK(!are_equivalent(smaller, bigger, &alph, params));
             CHECK(!are_equivalent(smaller, bigger, params));
@@ -1742,7 +1817,7 @@ TEST_CASE("Mata::Nfa::are_equivalent")
         bigger.final = {11};
 
         for (const auto& algo : ALGORITHMS) {
-            params["algo"] = algo;
+            params["algorithm"] = algo;
 
             CHECK(are_equivalent(smaller, bigger, &alph, params));
             CHECK(are_equivalent(smaller, bigger, params));
@@ -1768,7 +1843,7 @@ TEST_CASE("Mata::Nfa::are_equivalent")
         bigger.delta.add(11, alph["b"], 11);
 
         for (const auto& algo : ALGORITHMS) {
-            params["algo"] = algo;
+            params["algorithm"] = algo;
 
             //TODO:what about we test the plumbing versions primarily?
             // Debugging with the dispatcher is annoying.
@@ -1815,7 +1890,7 @@ TEST_CASE("Mata::Nfa::are_equivalent")
         bigger.delta.add(15, alph["b"], 15);
 
         for (const auto& algo : ALGORITHMS) {
-            params["algo"] = algo;
+            params["algorithm"] = algo;
 
             CHECK(!are_equivalent(smaller, bigger, &alph, params));
             CHECK(!are_equivalent(smaller, bigger, params));
@@ -1841,7 +1916,7 @@ TEST_CASE("Mata::Nfa::are_equivalent")
     SECTION("wrong parameters 2")
     {
         OnTheFlyAlphabet alph{};
-        params["algo"] = "foo";
+        params["algorithm"] = "foo";
 
         CHECK_THROWS_WITH(are_equivalent(smaller, bigger, &alph, params),
                           Catch::Contains("received an unknown value"));
@@ -2239,11 +2314,11 @@ TEST_CASE("Mata::Nfa::reduce_size_by_simulation()")
 
 	SECTION("empty automaton")
 	{
-		Nfa result = reduce(aut, &state_map);
+		Nfa result = reduce(aut, false, &state_map);
 
 		REQUIRE(nothing_in_trans(result));
-		REQUIRE(result.initial.size() == 0);
-		REQUIRE(result.final.size() == 0);
+		REQUIRE(result.initial.empty());
+		REQUIRE(result.final.empty());
 	}
 
 	SECTION("simple automaton")
@@ -2252,7 +2327,7 @@ TEST_CASE("Mata::Nfa::reduce_size_by_simulation()")
         aut.initial.add(1);
 
         aut.final.add(2);
-		Nfa result = reduce(aut, &state_map);
+		Nfa result = reduce(aut, false, &state_map);
 
 		REQUIRE(nothing_in_trans(result));
 		REQUIRE(result.initial[state_map[1]]);
@@ -2284,7 +2359,7 @@ TEST_CASE("Mata::Nfa::reduce_size_by_simulation()")
 		aut.final = {3, 9};
 
 
-		Nfa result = reduce(aut, &state_map);
+		Nfa result = reduce(aut, false, &state_map);
 
 		REQUIRE(result.size() == 6);
 		REQUIRE(result.initial[state_map[1]]);
@@ -2299,6 +2374,32 @@ TEST_CASE("Mata::Nfa::reduce_size_by_simulation()")
 		REQUIRE(result.delta.contains(state_map[2], 'a', state_map[2]));
 		REQUIRE(result.final[state_map[9]]);
 		REQUIRE(result.final[state_map[3]]);
+
+        result = reduce(aut, true, &state_map);
+        CHECK(result.size() == 3);
+        CHECK(result.initial.size() == 2);
+        for (State initial : result.initial) {
+            CHECK(((initial == state_map[1]) || (initial == state_map[2])));
+        }
+        REQUIRE(result.final.size() == 1);
+        for (State final : result.final) {
+            CHECK(final == state_map[3]);
+        }
+        CHECK(result.delta.size() == 6);
+        CHECK(result.delta.contains(state_map[1], 'a', state_map[3]));
+        CHECK(result.delta.contains(state_map[1], 'a', state_map[2]));
+        CHECK(result.delta.contains(state_map[2], 'a', state_map[2]));
+        CHECK(result.delta.contains(state_map[2], 'b', state_map[2]));
+        CHECK(result.delta.contains(state_map[2], 'a', state_map[3]));
+        CHECK(result.delta.contains(state_map[3], 'b', state_map[2]));
+	}
+
+	SECTION("no transitions from non-final state")
+	{
+		aut.delta.add(0, 'a', 1);
+		aut.initial = { 0 };
+		Nfa result = reduce(aut, &state_map);
+		CHECK(Mata::Nfa::are_equivalent(result, aut));
 	}
 }
 
@@ -2610,19 +2711,44 @@ TEST_CASE("Mata::Nfa::trim()")
 
     Nfa old_aut{aut};
 
-    aut.trim();
-    CHECK(aut.initial.size() == old_aut.initial.size());
-    CHECK(aut.final.size() == old_aut.final.size());
-    CHECK(aut.size() == 4);
-    for (const Word& word: get_shortest_words(old_aut))
-    {
-        CHECK(is_in_lang(aut, Run{word,{}}));
+    SECTION("Without state map") {
+        aut.trim();
+        CHECK(aut.initial.size() == old_aut.initial.size());
+        CHECK(aut.final.size() == old_aut.final.size());
+        CHECK(aut.size() == 4);
+        for (const Word& word: get_shortest_words(old_aut))
+        {
+            CHECK(is_in_lang(aut, Run{word,{}}));
+        }
+
+        aut.final.remove(2); // '2' is the new final state in the earlier trimmed automaton.
+        aut.trim();
+        CHECK(aut.delta.empty());
+        CHECK(aut.size() == 0);
     }
 
-    aut.final.remove(2); // '2' is the new final state in the earlier trimmed automaton.
-    aut.trim();
-    CHECK(aut.delta.empty());
-    CHECK(aut.size() == 0);
+    SECTION("With state map") {
+        StateToStateMap state_map{};
+        aut.trim(&state_map);
+        CHECK(aut.initial.size() == old_aut.initial.size());
+        CHECK(aut.final.size() == old_aut.final.size());
+        CHECK(aut.size() == 4);
+        for (const Word& word: get_shortest_words(old_aut))
+        {
+            CHECK(is_in_lang(aut, Run{word,{}}));
+        }
+        REQUIRE(state_map.size() == 4);
+        CHECK(state_map.at(1) == 0);
+        CHECK(state_map.at(3) == 1);
+        CHECK(state_map.at(7) == 3);
+        CHECK(state_map.at(5) == 2);
+
+        aut.final.remove(2); // '2' is the new final state in the earlier trimmed automaton.
+        aut.trim(&state_map);
+        CHECK(aut.delta.empty());
+        CHECK(aut.size() == 0);
+        CHECK(state_map.empty());
+    }
 }
 
 TEST_CASE("Mata::Nfa::Nfa::delta.empty()")
@@ -2681,18 +2807,25 @@ TEST_CASE("Mata::Nfa::delta.operator[]")
     FILL_WITH_AUT_A(aut);
     REQUIRE(aut.get_num_of_trans() == 15);
     aut.delta[25];
+    REQUIRE(aut.size() == 20);
 
+    aut.delta.get_mutable_post(25);
     REQUIRE(aut.size() == 26);
     REQUIRE(aut.delta[25].empty());
 
-    aut.delta[50];
+    aut.delta.get_mutable_post(50);
     REQUIRE(aut.size() == 51);
     REQUIRE(aut.delta[50].empty());
 
-    const Nfa aut1 = aut;
-    aut1.delta[60];
+    Nfa aut1 = aut;
+    aut1.delta.get_mutable_post(60);
     REQUIRE(aut1.size() == 61);
     REQUIRE(aut1.delta[60].empty());
+
+    const Nfa aut2 = aut;
+    aut2.delta[60];
+    REQUIRE(aut2.size() == 51);
+    REQUIRE(aut2.delta[60].empty());
 }
 
 TEST_CASE("Mata::Nfa::Nfa::unify_(initial/final)()") {
@@ -2856,14 +2989,26 @@ TEST_CASE("Mata::Nfa::Nfa::defragment()") {
     aut.delta.add(4, 42, 2);
     aut.delta.add(4, 42, 1);
     aut.delta.add(5, 42, 4);
+    aut.initial.add(3);
+    aut.final.add(3);
+    aut.final.add(4);
+    CHECK(aut.initial[3]);
+    CHECK(aut.final[3]);
+    CHECK(aut.final[4]);
     CHECK(aut.size() == 6);
     CHECK(aut.delta.contains(1, 42, 2));
     aut.defragment();
-    CHECK(aut.size() == 5);
+    CHECK(aut.initial[5]);
+    CHECK(aut.final[5]);
+    CHECK(aut.final[3]);
+    CHECK(!aut.final[4]);
+    CHECK(aut.size() == 6);
     CHECK(aut.delta.contains(1, 42, 3));
     CHECK(aut.delta.contains(4, 42, 3)); // previously (5,42,4)
     aut.defragment();
-    CHECK(aut.size() == 5);
+    CHECK(aut.initial[5]);
+    CHECK(aut.final[5]);
+    CHECK(aut.size() == 6);
     // transitions contains not been changed
     CHECK(aut.delta.contains(1, 42, 3));
     CHECK(aut.delta.contains(4, 42, 3));
@@ -2885,3 +3030,23 @@ TEST_CASE("A segmentation fault in the make_complement") {
     make_complete(r, alph, 1);
     REQUIRE(is_complete(r, alph));
 }
+
+TEST_CASE("Mata::Nfa:: create simple automata") {
+    Nfa nfa{ create_empty_string_nfa() };
+    CHECK(is_in_lang(nfa, { {}, {} }));
+    CHECK(get_word_lengths(nfa) == std::set<std::pair<int, int>>{ std::make_pair(0, 0) });
+
+    OnTheFlyAlphabet alphabet{};
+    StringToSymbolMap symbol_map{ { "a", 0 }, { "b", 1 }, { "c", 2 } };
+    alphabet.add_symbols_from(symbol_map);
+    nfa = create_sigma_star_nfa(&alphabet);
+    CHECK(is_in_lang(nfa, { {}, {} }));
+    CHECK(is_in_lang(nfa, { { 0 }, {} }));
+    CHECK(is_in_lang(nfa, { { 1 }, {} }));
+    CHECK(is_in_lang(nfa, { { 2 }, {} }));
+    CHECK(is_in_lang(nfa, { { 0, 1 }, {} }));
+    CHECK(is_in_lang(nfa, { { 1, 0 }, {} }));
+    CHECK(is_in_lang(nfa, { { 2, 2, 2 }, {} }));
+    CHECK(is_in_lang(nfa, { { 0, 1, 2, 2, 0, 1, 2, 1, 0, 0, 2, 1 }, {} }));
+    CHECK(!is_in_lang(nfa, { { 3 }, {} }));
+}
